"""
Functions to determine and setup parameters
"""

import numpy as np
import pandas as pd
from IPython.core.display import display
## TODO FIX yaml issues
## from ruamel import yaml


def hack_make_lines_from_csv(fn='../../../data/DIRECT_IEE.csv', do_save=False):
    """
    placeholder
    make industry lines from IEE extract
    provenance: IND_IEE*....py in python

    :param do_save:
    :param fn:
    :return:
    """
    D = pd.read_csv(fn)
    line_to_pers_comm = {'Aggregate write-ins for Other Lines of Business': 'Commercial',
                         'Aircraft (all perils)': 'Commercial',
                         'Allied Lines': 'Commercial',
                         'Boiler and Machinery': 'Commercial',
                         'Burglary and theft': 'Commercial',
                         'Commercial Auto Liability': 'Commercial',
                         'Commercial Auto Physical Damage': 'Commercial',
                         'Commercial Multiple Peril (Liability Portion)': 'Commercial',
                         'Commercial Multiple Peril (Non-Liability Portion)': 'Commercial',
                         'Credit': 'Commercial',
                         'Credit A & H': 'Commercial',
                         'Earthquake': 'Commercial',
                         "Excess workers' compensation": 'Commercial',
                         'Farmowners Multiple Peril': 'Personal',
                         'Federal Flood': 'Commercial',
                         'Fidelity': 'Commercial',
                         'Financial Guaranty': 'Commercial',
                         'Fire': 'Commercial',
                         'Group A&H (See Interrogatory 1)': 'Commercial',
                         'Homeowners Multiple Peril': 'Personal',
                         'Inland Marine': 'Commercial',
                         'International': 'Commercial',
                         'Medical Professional Liability': 'Commercial',
                         'Mortgage Guaranty': 'Commercial',
                         'Multiple Peril Crop': 'Commercial',
                         'Ocean Marine': 'Commercial',
                         'Other A&H (See Interrogatory 1)': 'Commercial',
                         'Other Liability - Claims-made': 'Commercial',
                         'Other Liability - Occurrence': 'Commercial',
                         'Private Crop': 'Commercial',
                         'Private Passenger Auto Liability': 'Personal',
                         'Private Passenger Auto Physical Damage': 'Personal',
                         'Products Liability': 'Commercial',
                         'Reinsurance-Nonproportional Assumed': 'Commercial',
                         'Surety': 'Commercial',
                         'TOTAL NET': 'DROP',
                         'TOTALS DIRECT and ASSUMED': 'DROP',
                         'TOTALS DIRECT and PROPORTIONAL ASSUMED (Lines 1 through 34)': 'DROP',
                         'Warranty': 'Commercial',
                         "Workers' Compensation": 'Commercial'}

    line_to_prop_auto_liab = {'Aggregate write-ins for Other Lines of Business': 'Liability',
                              'Aircraft (all perils)': 'Liability',
                              'Allied Lines': 'Property',
                              'Boiler and Machinery': 'Property',
                              'Burglary and theft': 'Property',
                              'Commercial Auto Liability': 'Auto',
                              'Commercial Auto Physical Damage': 'Auto',
                              'Commercial Multiple Peril (Liability Portion)': 'Liability',
                              'Commercial Multiple Peril (Non-Liability Portion)': 'Property',
                              'Credit': 'Liability',
                              'Credit A & H': 'Liability',
                              'Earthquake': 'Property',
                              "Excess workers' compensation": 'Liability',
                              'Farmowners Multiple Peril': 'Property',
                              'Federal Flood': 'Property',
                              'Fidelity': 'Liability',
                              'Financial Guaranty': 'Liability',
                              'Fire': 'Property',
                              'Group A&H (See Interrogatory 1)': 'Liability',
                              'Homeowners Multiple Peril': 'Property',
                              'Inland Marine': 'Property',
                              'International': 'Liability',
                              'Medical Professional Liability': 'Liability',
                              'Mortgage Guaranty': 'Liability',
                              'Multiple Peril Crop': 'Property',
                              'Ocean Marine': 'Liability',
                              'Other A&H (See Interrogatory 1)': 'Liability',
                              'Other Liability - Claims-made': 'Liability',
                              'Other Liability - Occurrence': 'Liability',
                              'Private Crop': 'Property',
                              'Private Passenger Auto Liability': 'Auto',
                              'Private Passenger Auto Physical Damage': 'Auto',
                              'Products Liability': 'Liability',
                              'Reinsurance-Nonproportional Assumed': 'Liability',
                              'Surety': 'Liability',
                              'TOTAL NET': 'DROP',
                              'TOTALS DIRECT and ASSUMED': 'DROP',
                              'TOTALS DIRECT and PROPORTIONAL ASSUMED (Lines 1 through 34)': 'DROP',
                              'Warranty': 'Liability',
                              "Workers' Compensation": 'Liability'}

    line_to_prop_liab = {'Aggregate write-ins for Other Lines of Business': 'Liability',
                         'Aircraft (all perils)': 'Liability',
                         'Allied Lines': 'Property',
                         'Boiler and Machinery': 'Property',
                         'Burglary and theft': 'Property',
                         'Commercial Auto Liability': 'Liability',
                         'Commercial Auto Physical Damage': 'Property',
                         'Commercial Multiple Peril (Liability Portion)': 'Liability',
                         'Commercial Multiple Peril (Non-Liability Portion)': 'Property',
                         'Credit': 'Liability',
                         'Credit A & H': 'Liability',
                         'Earthquake': 'Property',
                         "Excess workers' compensation": 'Liability',
                         'Farmowners Multiple Peril': 'Property',
                         'Federal Flood': 'Property',
                         'Fidelity': 'Liability',
                         'Financial Guaranty': 'Liability',
                         'Fire': 'Property',
                         'Group A&H (See Interrogatory 1)': 'Liability',
                         'Homeowners Multiple Peril': 'Property',
                         'Inland Marine': 'Property',
                         'International': 'Liability',
                         'Medical Professional Liability': 'Liability',
                         'Mortgage Guaranty': 'Liability',
                         'Multiple Peril Crop': 'Property',
                         'Ocean Marine': 'Liability',
                         'Other A&H (See Interrogatory 1)': 'Liability',
                         'Other Liability - Claims-made': 'Liability',
                         'Other Liability - Occurrence': 'Liability',
                         'Private Crop': 'Property',
                         'Private Passenger Auto Liability': 'Liability',
                         'Private Passenger Auto Physical Damage': 'Property',
                         'Products Liability': 'Liability',
                         'Reinsurance-Nonproportional Assumed': 'Liability',
                         'Surety': 'Liability',
                         'TOTAL NET': 'DROP',
                         'TOTALS DIRECT and ASSUMED': 'DROP',
                         'TOTALS DIRECT and PROPORTIONAL ASSUMED (Lines 1 through 34)': 'DROP',
                         'Warranty': 'Liability',
                         "Workers' Compensation": 'Liability'}

    D.loc[:, 'PER_COMM'] = D.FULL_NAME.map(line_to_pers_comm)
    D.loc[:, 'PROP_AUTO_LIAB'] = D.FULL_NAME.map(line_to_prop_auto_liab)
    D.loc[:, 'PROP_LIAB'] = D.FULL_NAME.map(line_to_prop_liab)

    # summarize across lines and including loss adj expense
    line_namer = {'Aggregate write-ins for Other Lines of Business': 'Commercial',
                  'Aircraft (all perils)': 'Commercial',
                  'Allied Lines': 'commprop',
                  'Boiler and Machinery': 'Commercial',
                  'Burglary and theft': 'Commercial',
                  'Commercial Auto Liability': 'cal',
                  'Commercial Auto Physical Damage': 'cal',
                  'Commercial Multiple Peril (Liability Portion)': 'cmp',
                  'Commercial Multiple Peril (Non-Liability Portion)': 'cmp',
                  'Credit': 'Commercial',
                  'Credit A & H': 'Commercial',
                  'Earthquake': 'Commercial',
                  "Excess workers' compensation": 'Commercial',
                  'Farmowners Multiple Peril': 'Personal',
                  'Federal Flood': 'Commercial',
                  'Fidelity': 'Commercial',
                  'Financial Guaranty': 'Commercial',
                  'Fire': 'commprop',
                  'Group A&H (See Interrogatory 1)': 'Commercial',
                  'Homeowners Multiple Peril': 'ho',
                  'Inland Marine': 'im',
                  'International': 'Commercial',
                  'Medical Professional Liability': 'med mal',
                  'Mortgage Guaranty': 'Commercial',
                  'Multiple Peril Crop': 'Commercial',
                  'Ocean Marine': 'Commercial',
                  'Other A&H (See Interrogatory 1)': 'Commercial',
                  'Other Liability - Claims-made': 'Commercial',
                  'Other Liability - Occurrence': 'Commercial',
                  'Private Crop': 'Commercial',
                  'Private Passenger Auto Liability': 'ppa',
                  'Private Passenger Auto Physical Damage': 'ppa',
                  'Products Liability': 'Commercial',
                  'Reinsurance-Nonproportional Assumed': 're assumed',
                  'Surety': 'Commercial',
                  'TOTAL NET': 'DROP',
                  'TOTALS DIRECT and ASSUMED': 'total',
                  'TOTALS DIRECT and PROPORTIONAL ASSUMED (Lines 1 through 34)': 'DROP',
                  'Warranty': 'Commercial',
                  "Workers' Compensation": 'wc'}

    def clean(D):
        D['DIL'] = (D.IL + D.DCC + D.AOE)
        D['E'] = pd.to_numeric(D.CUSTOMER, errors='coerce') + + pd.to_numeric(D.PAPER, errors='coerce')
        df = D[['FULL_NAME', 'PROP_AUTO_LIAB', 'YEAR', 'DEP', 'DIL', 'E']].copy()
        df = df.dropna()
        test_lines = ['Fire', 'Allied Lines',
                      'Homeowners Multiple Peril',
                      'Medical Professional Liability',
                      'Commercial Multiple Peril (Non-Liability Portion)',
                      'Commercial Multiple Peril (Liability Portion)',
                      'Inland Marine',
                      "Workers' Compensation",
                      'Private Passenger Auto Liability',
                      'Commercial Auto Liability',
                      'Private Passenger Auto Physical Damage',
                      'TOTALS DIRECT and ASSUMED']
        test_lines = [f'"{i}"' for i in test_lines]
        query = f' FULL_NAME in [{", ".join(test_lines)}]'
        test_df = df.query(query)
        test_df['NAME'] = test_df.FULL_NAME.map(line_namer)
        test_df = test_df.drop('FULL_NAME', axis=1)
        test_df = test_df.groupby(['NAME', 'YEAR']).sum()

        test_df['LR'] = test_df.DIL.divide(test_df.DEP)
        test_df['ER'] = test_df.E.divide(test_df.DEP)
        test_df['CR'] = test_df.LR + test_df.ER
        test_df['PLR'] = test_df.LR.divide(1 - test_df.ER)
        test_df = test_df.reset_index()
        return test_df

    def cov(x):
        return x.std() / x.mean()

    def skew(x):
        return x.skew()

    def premium_net_of_expenses(x):
        """
        premium net of expenses

        """
        return x.iloc[-1]

    sf = [np.mean, np.std, cov, skew]
    tl = clean(D)
    piv = tl.groupby('NAME').agg({'LR': sf, 'PLR': sf, 'ER': sf, 'CR': sf, 'DEP': premium_net_of_expenses})
    display(piv)
    piv.plot(kind='bar', subplots=True, figsize=(9, 8), layout=(-1, 4), sharex=True)
    ans = piv['DEP']
    ans['PLRCV'] = piv[('PLR', 'cov')]
    ans['ER'] = piv[('ER', 'mean')]
    ans['NEP'] = ans.EP * (1 - ans.ER)
    ans['esev'] = [40000, 100000, 250000, 15000, 50000, 750000, 15000, 15000, 10000]
    ans['cvsev'] = 0.5
    ans['limit'] = 3000000
    # ans['limit'] = [3000000, 2000000, 5000000, 1000000, 2000000, 500000, 10000000]

    d = dict()
    for name, l in ans.iterrows():
        s = dict(name='lognorm', mean=float(l['esev']), cv=float(l['cvsev']))
        f = dict(n=float(l.NEP * 1000 / l.esev), contagion=float(l.PLRCV ** 2))
        e = dict(name=name, severity=s, frequency=f)
        d[f'ind {name}'] = e
    print(yaml.dump(d, default_flow_style=False, indent=4))
    if do_save:
        with open('./agg/aggregate.yaml', 'a') as f:
            yaml.dump(d, stream=f, default_flow_style=False, indent=4)
    return tl



"""
Distortion functions to implement spectral risk measures

May 2019: added capped log linear and tt distortions
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as ss
from scipy.interpolate import interp1d
from scipy.spatial import ConvexHull
from io import StringIO
import pandas as pd
from . utils import axiter_factory, suptitle_and_tight


class Distortion(object):
    """
    handles everything to do with distortion functions


    """
    # make these (mostly) immutable...avoid changing by mistake
    _available_distortions_ = ('ph', 'wang', 'tt', 'cll', 'lep', 'ly', 'clin', 'dual', 'tvar', 'convex')
    _has_mass_ = ('ly', 'clin', 'lep')
    _med_names_ = ("Prop Hzrd", "Wang", 'Wang-tt', 'Capd Loglin', "Lev Equiv",
                    "Lin Yield", "Capped Linear", "Dual Mom", "Tail VaR", "Convex Env")
    _long_names_ = ("Proportional Hazard", "Wang-normal", 'Wang-tt', 'Capped Loglinear', "Leverage Equivalent Pricing",
                    "Linear Yield", "Capped Linear", "Dual Moment", "Tail VaR", "Convex Envelope")
    # TODO fix examples!
    # _available_distortions_ = ('ph', 'wang', 'tt', 'cll', 'lep',  'ly', 'clin', 'dual', 'tvar', 'convex')
    _eg_param_1_ =              (.9,     1,     1,     .9,    0.25,  0.9,   1.1, 3,  0.75)
    _eg_param_2_ =              (.5,     2,     2,     .8,    0.35,  1.5,   1.8, 6,  0.95)
    # _distortion_names_ = dict(zip(_available_distortions_, _med_names_))
    _distortion_names_ = dict(zip(_available_distortions_, _long_names_))

    @classmethod
    def available_distortions(cls, pricing=True, strict=True):
        """
        list of the available distortions

        :param pricing: only return list suitable for pricing, excludes tvar and convex
        :param strict: only include those without mass at zero  (pricing only)
        :return:
        """

        if pricing and strict:
            return tuple((i for i in cls._available_distortions_[:-2] if i not in cls._has_mass_))
        elif pricing:
            return cls._available_distortions_[:-2]
        else:
            return cls._available_distortions_

    def __init__(self, name, shape, r0=0.0, df=None, col_x='', col_y=''):
        """
        create new distortion

        Tester:

            ps = np.linspace(0, 1, 201)
            for dn in agg.Distortion.available_distortions(True):
                if dn=='clin':
                    # shape param must be > 1
                    g_dist = agg.Distortion(**{'name': dn, 'shape': 1.25, 'r0': 0.02, 'df': 5.5})
                else:
                    g_dist = agg.Distortion(**{'name': dn, 'shape': 0.5, 'r0': 0.02, 'df': 5.5})
                g_dist.plot()
                g = g_dist.g
                g_inv = g_dist.g_inv

                df = pd.DataFrame({'p': ps, 'gg_inv': g(g_inv(ps)), 'g_invg': g_inv(g(ps)),
                'g': g(ps), 'g_inv': g_inv(ps)})
                print(dn)
                print("errors")
                display(df.query(' abs(gg_inv - g_invg) > 1e-5'))


        :param name: name of an available distortion, call ``Distortion.available_distortions()`` for a list
        :param shape: float or [float, float]
        :param shape: shape parameter
        :param r0: risk free or rental rate of interest
        :param df:  for convex envelope, dataframe with col_x and col_y used to parameterize or df for t
        :param col_x:
        :param col_y:
        """
        self.name = name
        self.shape = shape
        self.r0 = r0
        # when created by calibrate distortions extra info put here
        self.error = 0.0
        self.premium_target = 0.0
        self.assets = 0.0
        self.mass = 0.0
        self.df = df
        self.col_x = col_x
        self.col_y = col_y

        # now make g and g_inv
        if self.name == 'ph':
            rho = self.shape
            rhoinv = 1.0 / rho
            self.has_mass = False

            # @numba.vectorize(["float64(float64)"], nopython=True, target='parallel')
            def g(x):
                return x ** rho

            def g_inv(x):
                return x ** rhoinv

        elif self.name == 'wang':
            lam = self.shape
            n = ss.norm()
            self.has_mass = False

            def g(x):
                return n.cdf(n.ppf(x) + lam)

            def g_inv(x):
                return n.cdf(n.ppf(x) - lam)

        elif self.name == 'tt':
            lam = self.shape
            t = ss.t(self.df)
            self.has_mass = False

            def g(x):
                return t.cdf(t.ppf(x) + lam)

            def g_inv(x):
                return t.cdf(t.ppf(x) - lam)

        elif self.name == 'cll':
            # capped log linear
            b = self.shape
            binv = 1 / b
            ea = np.exp(self.r0)
            a = self.r0
            self.has_mass = False

            def g(x):
                return np.where(x==0, 0, np.minimum(1, ea * x ** b))

            def g_inv(x):
                return np.where(x < 1, np.minimum(1, (x / ea) ** binv), 1)

        elif self.name == 'tvar':
            p = self.shape
            alpha = 1 / (1 - p)
            self.has_mass = False

            def g(x):
                return np.minimum(alpha * x, 1)

            def g_inv(x):
                return np.where(x < 1, x * (1 - p), 1)

        elif self.name == 'ly':
            # linear yield
            # r0 = occupancy; rk = consumption specified in list shape parameter
            rk = self.shape
            self.has_mass = (r0 > 0)
            self.mass = r0 / (1 + r0)

            def g(x):
                return np.where(x == 0, 0, (self.r0 + x * (1 + rk)) / (1 + self.r0 + rk * x))

            def g_inv(x):
                return np.maximum(0, (x * (1 + self.r0) - self.r0) / (1 + rk * (1 - x)))

        elif self.name == 'clin':
            # capped linear, needs shape > 1 to make sense...WHAT, WHY? needs shape >= 1-r0 else
            # problems at 1
            sl = self.shape
            self.has_mass = (r0 > 0)
            self.mass = r0

            def g(x):
                return np.where(x == 0, 0, np.minimum(1, self.r0 + sl * x))

            def g_inv(x):
                return np.where(x <= self.r0, 0, (x - self.r0) / sl)

        elif self.name == 'lep':
            # leverage equivalent pricing
            # self.r0 = risk free/financing and r = risk charge (the solved parameter)
            r = self.shape
            delta = r / (1 + r)
            d = self.r0 / (1 + self.r0)
            spread = delta - d
            self.has_mass = (d > 0)
            self.mass = d

            def g(x):
                return np.where(x == 0, 0, np.minimum(1, d + (1 - d) * x + spread * np.sqrt(x * (1 - x))))

            spread2 = spread ** 2
            a = (1 - d) ** 2 + spread2

            def g_inv(y):
                mb = (2 * (y - d) * (1 - d) + spread2)  # mb = -b
                c = (y - d) ** 2
                rad = np.sqrt(mb * mb - 4 * a * c)
                # l = (mb + rad)/(2 * a)
                u = (mb - rad) / (2 * a)
                return np.where(y < d, 0, np.maximum(0, u))

        elif self.name == 'dual':
            # dual moment
            p = self.shape
            q = 1 / p
            self.has_mass = False

            def g(x):
                return 1 - (1 - x)**p

            def g_inv(y):
                return 1 - (1 - y)**q

        elif self.name == 'convex':
            self.has_mass = False
            hull = ConvexHull(df[[col_x, col_y]])
            knots = list(set(hull.simplices.flatten()))
            g = interp1d(df.iloc[knots, df.columns.get_loc(col_x)],
                         df.iloc[knots, df.columns.get_loc(col_y)], kind='linear')
            g_inv = interp1d(df.iloc[knots, df.columns.get_loc(col_y)],
                         df.iloc[knots, df.columns.get_loc(col_x)], kind='linear')
        else:
            raise ValueError(
                "Incorrect spec passed to Distortion; implemented g types are ph, wang, tvar, "
                "ly (linear yield), lep (layer equivalent pricing) and clin (clipped linear)")

        self.g = g
        self.g_inv = g_inv
        def g_dual(x):
            return 1 - self.g(x)
        self.g_dual = g_dual

    def __str__(self):
        """
        printable version of distortion

        :return:
        """
        if isinstance(self.shape, str):
            s = f'{self._distortion_names_[self.name]}\n{self.shape}'
        else:
            s = f'{self._distortion_names_[self.name]}\n{self.shape:.3f}'
        if self.has_mass:
            s += f', {self.r0:.3f}'
        if self.name == 'tt':
            s += f', {self.df:.2f}'
        return s

    def __repr__(self):
        s = f'{self.name} ({self.shape}'
        if self.has_mass:
            s += f', {self.r0})'
        elif self.name == 'tt':
            s += f', {self.df:.2f})'
        else:
            s += ')'
        return s

    # noinspection PyUnusedLocal
    def plot(self, xs=None, n=101, both=True, ax=None, **kwargs):
        """
        quick plot of the distortion

        :param ax:
        :param xs:
        :param n:  length of vector is no xs
        :param both: ignored for now. just do both
        :param kwargs:  passed to plot
        :return:
        """

        if xs is None:
            xs = np.linspace(0, 1, n)

        y1 = self.g(xs)
        y2 = self.g_inv(xs)

        if ax is None:
            ax = plt.gca()

        ax.plot(xs, y1, label='$g$', **kwargs)
        ax.plot(xs, y2, label='$g^{-1}$', **kwargs)
        ax.plot(xs, xs, lw=0.5, color='black', alpha=0.5)
        if self.name == 'convex':
            ax.plot(self.df.loc[:, self.col_x], self.df.loc[:, self.col_y], 'x')
        ax.grid(which='major', axis='both', linestyle='-', linewidth='0.1', color='blue', alpha=0.5)
        ax.set(title=str(self), aspect='equal')
        return ax

    @classmethod
    def test(cls, r0=0.05):
        """
        tester: make some nice plots
        TODO add new distortions and fix df for tt

        :return:
        """

        axiter = axiter_factory(None, 18, figsize=(10, 10 * 4 / 5))

        xs = np.linspace(0, 1, 1001)

        # zip stops at the shorter of the vectors, so this does not include convex (must be listed last)
        # added df for the t; everyone else can ignore it
        for name, shape in zip(cls._available_distortions_, cls._eg_param_1_):
            dist = Distortion(name, shape, r0, df=4)
            dist.plot(xs, ax=next(axiter))

        dist = Distortion.convex_example('bond')
        dist.plot(xs, ax=next(axiter))

        dist = Distortion.convex_example('cat')
        dist.plot(xs, ax=next(axiter))

        # order will look better like this
        for name, shape in zip(cls._available_distortions_, cls._eg_param_2_):
            dist = Distortion(name, shape, r0, df=2)
            dist.plot(xs, ax=next(axiter))

        axiter.tidy()
        suptitle_and_tight('Example Distortion Functions')

    @staticmethod
    def distortions_from_params(params, index, r0=0.025, df=5.5, plot=True, axiter=None, pricing=True, strict=True):
        """
        make set of dist funs and inverses from params, output of port.calibrate_distortions
        params must just have one row for each method and be in the output format of cal_dist

        :param plot:
        :param index:
        :param r0: min rol parameters
        :param params: dataframe such that params[index, :] has a [lep, param] etc.
        pricing=True, strict=True: which distortions to allow
        df for t distribution
        :return:
        """
        temp = params.loc[index, :]
        dists = {}
        for dn in Distortion.available_distortions(pricing=pricing, strict=strict):
            param = float(temp.loc[dn, 'param'])
            dists[dn] = Distortion(name=dn, shape=param, r0=r0, df=df)

        if plot:
            axiter = axiter_factory(axiter, len(dists))
            # f, axs = plt.subplots(2, 3, figsize=(8, 6))
            # it = iter(axs.flatten())
            for dn in Distortion.available_distortions(pricing=pricing, strict=strict):
                dists[dn].plot(ax=next(axiter))
            try:
                axiter.tidy()
                plt.tight_layout()
            except:
                # fails if axiter is just an iteration of axis elements
                # assume then that constrained_layout =True so no tight layout
                pass

        return dists  # [g_lep, g_ph, g_wang, g_ly, g_clin]

    @staticmethod
    def convex_example(source='bond'):
        """
        example convex distortion using data from https://www.bis.org/publ/qtrpdf/r_qt0312e.pdf

        :param source: bond gives a bond yield curve example, cat gives cat bond / cat reinsurance pricing based example
        :return:
        """

        if source == 'bond':
            yield_curve = '''
            AAAA    0.000000  0.000000
            AAA     0.000018  0.006386
            AA      0.000144  0.007122
            A       0.000278  0.010291
            BBB     0.002012  0.017089
            BB      0.012674  0.036455
            B       0.040052  0.069181
            Z       1.000000  1.000000'''

            df = pd.read_fwf(StringIO(yield_curve))
            df.columns = ['Rating', 'EL', 'Yield']
            return Distortion('convex', 'Yield Curve', df=df, col_x='EL', col_y='Yield')

        elif source.lower() == 'cat':
            cat_bond = '''EL,ROL
            0.116196,0.32613
            0.088113,0.2452
            0.074811,0.22769
            0.056385,0.17131
            0.046923,0.15326
            0.032961,0.12222
            0.02807,0.11037
            0.024205,0.1022
            0.011564,0.07284
            0.005813,0.06004
            0,0
            1,1'''
            df = pd.read_csv(StringIO(cat_bond))
            return Distortion('convex', 'Cat Bond', df=df, col_x='EL', col_y='ROL')

        else:
            raise ValueError(f'Inadmissible value {source} passed to convex_example, expected yield or cat')



"""
=================
Underwriter Class
=================

The Underwriter is an easy to use interface into the computational functionality of aggregate.

The Underwriter
---------------

* Maintains a default library of severity curves
* Maintains a default library of aggregate distributions corresponding to industry losses in
  major classes of business, total catastrophe losses from major perils, and other useful constructs
* Maintains a default library of portfolios, including several example instances and examples used in
  papers on risk theory (e.g. the Bodoff examples)


The library functions can be listed using

::

        uw.list()

or, for more detail

::

        uw.describe()

A given example can be inspected using ``uw['cmp']`` which returns the defintion of the database
object cmp (an aggregate representing industry losses from the line Commercial Multiperil). It can
be created as an Aggregate class using ``ag = uw('cmp')``. The Aggregate class can then be updated,
plotted and various reports run on it. In iPython or Jupyter ``ag`` returns an informative HTML
description.

The real power of Underwriter is access to the agg scripting language (see parser module). The scripting
language allows severities, aggregates and portfolios to be created using more-or-less natural language.
For example

::

        pf = uw('''
        port MyCompanyBook
            agg LineA 100 claims 100000 xs 0 sev lognorm 30000 cv 1.25
            agg LineB 150 claims 250000 xs 5000 sev lognorm 50000 cv 0.9
            agg Cat 2 claims 100000000 xs 0 sev 500000 * pareto 1.8 - 500000
        ''')

creates a portfolio with three sublines, LineA, LineB and Cat. LineA is 100 (expected) claims, each pulled
from a lognormal distribution with mean of 30000 and coefficient of variation 1.25 within the layer
100000 xs 0 (i.e. limited at 100000). The frequency distribution is Poisson. LineB is similar. Cat is jsut
2 claims from the indicated limit, with severity given by a Pareto distribution with shape parameter 1.8,
scale 500000, shifted left by 500000. This corresponds to the usual Pareto with survival function
S(x) = (lambda / (lambda + x))^1.8, x >= 0.

The portfolio can be approximated using FFTs to convolve the aggregates and add the lines. The severities
are first discretized using a certain bucket-size (bs). The port object has a port.recommend_bucket() to
suggest reasonable buckets:

>> pf.recommend_bucket()

+-------+---------+--------+--------+--------+-------+-------+-------+------+------+
|       | bs10    | bs11   | bs12   | bs13   | bs14  | bs15  | bs16  | bs18 | bs20 |
+=======+=========+========+========+========+=======+=======+=======+======+======+
| LineA | 3,903   | 1,951  | 976    | 488    | 244   | 122   | 61.0  | 15.2 | 3.8  |
+-------+---------+--------+--------+--------+-------+-------+-------+------+------+
| LineB | 8,983   | 4,491  | 2,245  | 1,122  | 561   | 280   | 140   | 35.1 | 8.8  |
+-------+---------+--------+--------+--------+-------+-------+-------+------+------+
| Cat   | 97,656  | 48,828 | 24,414 | 12,207 | 6,103 | 3,051 | 1,525 | 381  | 95.4 |
+-------+---------+--------+--------+--------+-------+-------+-------+------+------+
| total | 110,543 | 55,271 | 27,635 | 13,817 | 6,908 | 3,454 | 1,727 | 431  | 108  |
+-------+---------+--------+--------+--------+-------+-------+-------+------+------+

The column bsNcorrespond to discretizing with 2**N buckets. The rows show suggested bucket sizes for each
line and in total. For example with N=13 (i.e. 8196 buckets) the suggestion is 13817. It is best the bucket
size is a divisor of any limits or attachment points, so we select 10000.

Updating can then be run as

::

    bs = 10000
    pf.update(13, bs)
    pf.report('quick')
    pf.plot('density')
    pf.plot('density', logy=True)
    print(pf)

    Portfolio name           MyCompanyBook
    Theoretic expected loss     10,684,541.2
    Actual expected loss        10,657,381.1
    Error                          -0.002542
    Discretization size                   13
    Bucket size                     10000.00
    <aggregate.port.Portfolio object at 0x0000023950683CF8>


Etc. etc.

"""

import os
import numpy as np
from IPython.core.display import display
import logging
import pandas as pd
from collections import Iterable
from .port import Portfolio
from .utils import html_title
from .distr import Aggregate, Severity
from .parser import UnderwritingLexer, UnderwritingParser
import re
import warnings

logger = logging.getLogger('aggregate')

class Underwriter(object):
    """
    The underwriter class constructs real world examples from stored and user input Lines and Accounts.
    Whereas Examples only produces simple Portfolios and Books, the Underwriter class is more flexible.

    Handles persistence
    Is interface into program parser
    Handles safe lookup from database for parser

    Persisitence to and from agg files

    """

    data_types = ['portfolio', 'aggregate', 'severity']

    def __init__(self, dir_name="", name='Rory', databases=[], glob=None, store_mode=True, update=False,
                 verbose=False, log2=10, debug=False, create_all=False):
        """

        :param dir_name:
        :param name:
        :param databases: set equal to None to load the default databases. Faster to load without them.
        :param glob: reference, e.g. to globals(), used to resolve meta.XX references
        :param store_mode: add newly created aggregates to the database?
        :param update:
        :param verbose:
        :param log2:
        :param debug: run parser in debug mode
        :param create_all: by default write only creates portfolios.
        """

        self.last_spec = None
        self.name = name
        self.update = update
        self.log2 = log2
        self.debug = debug
        self.verbose = verbose  # for update
        self.glob = glob
        self.lexer = UnderwritingLexer()
        self.parser = UnderwritingParser(self._safe_lookup, debug)
        # otherwise these are hidden from pyCharm....
        self.severity = {}
        self.aggregate = {}
        self.portfolio = {}
        if databases is None:
            databases = ['site.agg', 'user.agg']
        self.dir_name = dir_name
        if self.dir_name == '':
            self.dir_name = os.path.split(__file__)[0]
            self.dir_name = os.path.join(self.dir_name, 'agg')
        # make sure all database entries are stored:
        self.store_mode = True
        for fn in databases:
            with open(os.path.join(self.dir_name, fn), 'r') as f:
                program = f.read()
            # read in, parse, save to sev/agg/port dictionaries
            self._runner(program)
        # set desired store_mode
        self.store_mode = store_mode
        self.create_all = create_all

    def __getitem__(self, item):
        """
        handles self[item]

        subscriptable: try user portfolios, b/in portfolios, line, severity
        to access specifically use severity or line methods

        :param item:
        :return:
        """
        # much less fancy version:
        obj = self.portfolio.get(item, None)
        if obj is not None:
            logger.info(f'Underwriter.__getitem__ | found {item} of type port')
            return 'port', obj
        obj = self.aggregate.get(item, None)
        if obj is not None:
            logger.info(f'Underwriter.__getitem__ | found {item} of type agg')
            return 'agg', obj
        obj = self.severity.get(item, None)
        if obj is not None:
            logger.info(f'Underwriter.__getitem__ | found {item} of type sev')
            return 'sev', obj
        raise LookupError(f'Item {item} not found in any database')
        # old, clever, generic, unreadable, unnecessary generality
        # for k in self.databases.keys():
        #     if item in self.__getattribute__(k).keys():
        #         # stip the s off the name: Books to Book etc.
        #         return k, self.__getattribute__(k)[item]
        # raise LookupError

    def _repr_html_(self):
        s = [f'<h1>Underwriter {self.name}</h1>']
        s.append(
            f'Underwriter expert in all classes including {len(self.severity)} severities, {len(self.aggregate)} aggregates'
            f' and {len(self.portfolio)} portfolios<br>')
        for what in ['severity', 'aggregate', 'portfolio']:
            s.append(f'<b>{what.title()}</b>: ')
            s.append(', '.join([k for k in sorted(getattr(self, what).keys())]))
            s.append('<br>')
        s.append(f'<h3>Settings</h3>')
        for k in ['update', 'log2', 'store_mode', 'verbose', 'last_spec', 'create_all']:
            s.append(f'<span style="color: red;">{k}</span>: {getattr(self, k)}; ')
        return '\n'.join(s)

    # this is evil: it passes unknown things through to write...
    # def __getattr__(self, item):
    #     """
    #     handles self.item and returns an appropriate object
    #
    #     :param item:
    #     :return:
    #     """
    #     # print(f'Underwriter.__getattr__({item}) called')
    #     if item[0] == '_':
    #         # deal with the _ipython_canary_method_should_not_exist_
    #         # print('bailing')
    #         return
    #     else:
    #         return self.write(item)

    def __call__(self, portfolio_program, **kwargs):
        """
        make the Underwriter object callable; pass through to write

        :param portfolio_program:
        :return:
        """
        return self.write(portfolio_program, **kwargs)

    def list(self):
        """
        list all available databases

        :return:
        """
        sers = dict()
        for k in Underwriter.data_types:
            d = sorted(list(self.__getattribute__(k).keys()))
            sers[k.title()] = pd.Series(d, index=range(len(d)), name=k)
        df = pd.DataFrame(data=sers)
        # df.index.name = 'No.'
        df = df.fillna('')
        return df

    def describe(self, item_type='', pretty_print=False):
        """
        more informative version of list
        Pull notes for type items

        :return:
        """

        def deal_with_sequences(x):
            """
            pandas can't have a field set as a sequence
            need to check if x is a sequence and if so return something suitable...

            :param x:
            :return:
            """
            if isinstance(x, Iterable):
                return str(x)
            else:
                return x

        df = pd.DataFrame(columns=['Name', 'Type', 'Severity', 'ESev', 'Sev_a', 'Sev_b',
                                   'EN', 'Freq_a',
                                   'ELoss', 'Notes'])
        df = df.set_index('Name')
        df['ELoss'] = np.maximum(df.ELoss, df.ESev * df.EN)
        if item_type == '':
            item_type = Underwriter.data_types
        else:
            item_type = [item_type.lower()]
        for k in item_type:  # self.databases.keys():
            for kk, vv in self.__getattribute__(k).items():
                _data_fields = [vv.get('sev_name', ''), vv.get('sev_mean', 0), vv.get('sev_a', 0),
                                vv.get('sev_b', 0), vv.get('exp_en', 0), vv.get('freq_a', 0),
                                vv.get('exp_el', 0), vv.get('note', '')]
                try:
                    df.loc[kk, :] = [k] + _data_fields
                except ValueError as e:
                    if e.args[0] == "setting an array element with a sequence":
                        df.loc[kk, :] = [k] + list(map(deal_with_sequences, _data_fields))
                    else:
                        raise e
        df = df.fillna('')
        if pretty_print:
            for t, egs in df.groupby('Type'):
                html_title(t, 2)
                display(egs.style)
        return df

    def parse_portfolio_program(self, portfolio_program, output='spec'):
        """
        Utility routine to parse the program and return the spec suitable to pass to Portfolio to
        create the object.
        Initially just for a single portfolio program (which it checks!)
        No argument of default conniptions

        To write program in testing mode use output='df':

        * dictionary definitions are added to uw but no objects are created
        * returns data frame description of added severity/aggregate/portfolios
        * the dataframe of aggregates can be used to create a portfolio (with all the aggregates) by calling

        ```Portfolio.from_DataFrame(name df)```

        To parse and get dictionary definitions use output='spec'.
        Aggregate and severity objects are also returned though they could be
        accessed directly using wu['name']. May be convenient...we'll see.

        Output has form that an Aggregate can be created from Aggregate(**x['name'])
        etc. which is a bit easier than uw['name'] which returns the type.

        TODO make more robust

        :param portfolio_program:
        :param output:  'spec' output a spec (assumes only one portfolio),
                        or a dictionary {name: spec_list} if multiple
                        'df' or 'dataframe' output as pandas data frame
                        'dict' output as dictionary of pandas data frames (old write_test output)
        :return:
        """

        self._runner(portfolio_program)

        # if globs replace all meta objects with a lookup object
        # copy from code below FRAGILE
        if self.glob is not None:
            for a in list(self.parser.agg_out_dict.values()) + list(self.parser.sev_out_dict.values()):
                if a['sev_name'][0:4] == 'meta':
                    obj_name = a['sev_name'][5:]
                    try:
                        obj = self.glob[obj_name]
                    except NameError as e:
                        print(f'Object {obj_name} passed as a proto-severity cannot be found')
                        raise e
                    a['sev_name'] = obj
                    logger.info(f'Underwriter.write | {a["sev_name"]} ({type(a)} reference to {obj_name} '
                                 f'replaced with object {obj.name} from glob')

        if output == 'spec':
            # expecting a single portfolio for this simple function
            # create the spec list string
            if len(self.parser.port_out_dict) == 1:
                # this behaviour to ensure backwards compatibility
                nm = ""
                spec_list = None
                for nm in self.parser.port_out_dict.keys():
                    # remember the spec comes back as a list of aggs that have been entered into the uw
                    # self[v] = ('agg', dictionary def) of the agg component v of the portfolio
                    spec_list = [self[v][1] for v in self.portfolio[nm]['spec']]
                return nm, spec_list

            elif len(self.parser.port_out_dict) > 1 or \
                    len(self.parser.agg_out_dict) or len(self.parser.sev_out_dict):
                # return dictionary: {pf_name : { name: pf_name, spec_list : [list] }}
                # so that you can call Portfolio(*output[pf_name]) to create pf_name
                # notes are dropped...
                ans = {}
                for nm in self.parser.port_out_dict.keys():
                    # remember the spec comes back as a list of aggs that have been entered into the uw
                    # self[v] = ('agg', dictionary def) of the agg component v of the portfolio
                    spec_list = [self[v][1] for v in self.portfolio[nm]['spec']]
                    ans[nm] = dict(name=nm, spec_list=spec_list)

                for nm in self.parser.agg_out_dict.keys():
                    ans[nm] = self.aggregate[nm]

                for nm in self.parser.sev_out_dict.keys():
                    ans[nm] = self.severity[nm]

                return ans

            else:
                logger.warning(f'Underwriter.parse_portfolio_program | program has no Portfolio outputs. '
                                'Nothing returned. ')
                return

        elif output == 'df' or output.lower() == 'dataframe':
            logger.info(f'Runner.write_test | Executing program\n{portfolio_program[:500]}\n\n')
            ans = {}
            if len(self.parser.sev_out_dict) > 0:
                for v in self.parser.sev_out_dict.values():
                    Underwriter._add_defaults(v, 'sev')
                ans['sev'] = pd.DataFrame(list(self.parser.sev_out_dict.values()),
                                          index=self.parser.sev_out_dict.keys())
            if len(self.parser.agg_out_dict) > 0:
                for v in self.parser.agg_out_dict.values():
                    Underwriter._add_defaults(v)
                ans['agg'] = pd.DataFrame(list(self.parser.agg_out_dict.values()),
                                          index=self.parser.agg_out_dict.keys())
            if len(self.parser.port_out_dict) > 0:
                ans['port'] = pd.DataFrame(list(self.parser.port_out_dict.values()),
                                           index=self.parser.port_out_dict.keys())
            return ans

        else:
            raise ValueError(f'Inadmissible output type {output}  passed to parse_portfolio_program. '
                             'Expecting spec or df/dataframe.')

    def write(self, portfolio_program, **kwargs):
        """
        Write a natural language program. Write carries out the following steps.

        1. Read in the program and cleans it (e.g. punctuation, parens etc. are
        removed and ignored, replace ; with new line etc.)
        2. Parse line by line to create a dictioonary definition of sev, agg or port objects
        3. If glob set, pull in objects
        4. replace sev.name, agg.name and port.name references with their objects
        5. If create_all set, create all objects and return in dictionary. If not set only create the port objects
        6. If update set, update all created objects.

        Sample input

        ::

            port MY_PORTFOLIO
                agg Line1 20  loss 3 x 2 sev gamma 5 cv 0.30 mixed gamma 0.4
                agg Line2 10  claims 3 x 2 sevgamma 12 cv 0.30 mixed gamma 1.2
                agg Line 3100  premium at 0.4 3 x 2 sev 4 * lognormal 3 cv 0.8 fixed 1

        The indents are required...

        See parser for full language spec! See Aggregate class for many examples.

        Reasonable kwargs:

        * bs
        * log2
        * verbose
        * update overrides class default
        * add_exa should port.add_exa add the exa related columns to the output?
        * create_all: create all objects, default just portfolios. You generally
                     don't want to create underlying sevs and aggs in a portfolio.

        :param portfolio_program:
        :param kwargs:
        :return: single created object or dictionary name: object
        """

        # prepare for update
        # what / how to do; little awkward: to make easier for user have to strip named update args
        # out of kwargs
        create_all = kwargs.get('create_all', self.create_all)
        update = kwargs.get('update', self.update)
        if update:
            if 'log2' in kwargs:
                log2 = kwargs.get('log2')
                del kwargs['log2']
            else:
                log2 = self.log2
            if 'bs' in kwargs:
                bs = kwargs.get('bs')
                del kwargs['bs']
            else:
                bs = 0
            if 'verbose' in kwargs:
                verbose = kwargs.get('verbose')
                del kwargs['verbose']
            else:
                verbose = self.verbose
            if 'add_exa' in kwargs:
                add_exa = kwargs.get('add_exa')
                del kwargs['add_exa']
            else:
                add_exa = False

        # function to handle update madness, use in either script or lookup updats for ports
        def _update(s, k):
            if update:
                if bs > 0 and log2 > 0:
                    _bs = bs
                    _log2 = log2
                else:
                    if bs == 0:  # and log2 > 0
                        # for log2 = 10
                        _bs = s.recommend_bucket().iloc[-1, 0]
                        _log2 = log2  # which must be > 0
                        # adjust bucket size for new actual log2
                        _bs *= 2 ** (10 - _log2)
                    else:  # bs > 0 and log2 = 0 which doesn't really make sense...
                        logger.warning('Underwriter.write | nonsensical options bs > 0 and log2 = 0')
                        _bs = bs
                        _log2 = 10
                logger.info(f"Underwriter.write | updating Portfolio {k} log2={_log2}, bs={_bs}")
                s.update(log2=_log2, bs=_bs, verbose=verbose, add_exa=add_exa, **kwargs)

        # first see if it is a built in object
        lookup_success = True
        _type = ''
        obj = None
        try:
            _type, obj = self.__getitem__(portfolio_program)
        except LookupError:
            lookup_success = False
            logger.info(f'underwriter.write | object {portfolio_program[:500]} not found, will process as program')
        if lookup_success:
            logger.info(f'underwriter.write | object {portfolio_program[:500]} found, returning object...')
            if _type == 'agg':
                # TODO, sure this isn't the solution to the double name problem....
                _name = obj.get('name', portfolio_program)
                obj = Aggregate(_name, **{k: v for k, v in obj.items() if k != 'name'})
                if update:
                    obj.easy_update(log2, bs)
                return obj
            elif _type == 'port':
                # actually make the object
                obj = Portfolio(portfolio_program, [self[v][1] for v in obj['spec']])
                _update(obj, portfolio_program)
                return obj
            elif _type == 'sev':
                if 'sev_wt' in obj:
                    del obj['sev_wt']
                return Severity(**obj)
            else:
                ValueError(f'Cannot build {_type} objects')
            return obj

        # if you fall through to here then the portfolio_program did not refer to a built in object
        # run the program
        self._runner(portfolio_program)

        # if globs replace all meta objects with a lookup object
        if self.glob is not None:
            logger.info(f'Underwriter.write | Resolving globals')
            for a in list(self.parser.agg_out_dict.values()) + list(self.parser.sev_out_dict.values()):
                if a['sev_name'][0:4] == 'meta':
                    logger.info(f'Underwriter.write | Resolving {a["sev_name"]}')
                    obj_name = a['sev_name'][5:]
                    try:
                        obj = self.glob[obj_name]
                    except NameError as e:
                        print(f'Object {obj_name} passed as a proto-severity cannot be found')
                        raise e
                    a['sev_name'] = obj
                    logger.info(f'Underwriter.write | {a["sev_name"]} ({type(a)} reference to {obj_name} '
                                 f'replaced with object {obj.name} from glob')
            logger.info(f'Underwriter.write | Done resolving globals')

        # create objects
        # 2019-11: create all objects not just the portfolios if create_all==True
        rv = None
        if len(self.parser.port_out_dict) > 0:
            # create ports
            rv = {}
            for k in self.parser.port_out_dict.keys():
                # remember the spec comes back as a list of aggs that have been entered into the uw
                s = Portfolio(k, [self[v][1] for v in self.portfolio[k]['spec']])
                s.program = 'unknown'
                _update(s, k)
                rv[k] = s
            if len(self.parser.port_out_dict) == 1:
                # only one portfolio so we can set its program
                s.program = portfolio_program

        if len(self.parser.agg_out_dict) > 0 and create_all:
            # new aggs, create them
            if rv is None:
                rv = {}
            for k, v in self.parser.agg_out_dict.items():
                # TODO FIX this clusterfuck
                s = Aggregate(k, **{kk: vv for kk, vv in v.items() if kk != 'name'})
                if update:
                    s.easy_update(self.log2, verbose=verbose)
                rv[k] = s

        if len(self.parser.sev_out_dict) > 0 and create_all:
            # new sevs, create them
            if rv is None:
                rv = {}
            for v in self.parser.sev_out_dict.values():
                if 'sev_wt' in v:
                    del v['sev_wt']
                s = Severity(**v)
                rv[f'sev_{s.__repr__()[38:54]}'] = s

        # report on what has been done
        if rv is None:
            print('WARNING: Program did not contain any output...')
            logger.warning(f'Underwriter.write | Program {portfolio_program} did not contain any output')
        else:
            if len(rv):
                logger.info(f'Underwriter.write | Program created {len(rv)} objects and '
                             f'defined {len(self.parser.port_out_dict)} Portfolio(s), '
                             f'{len(self.parser.agg_out_dict)} Aggregate(s), and '
                             f'{len(self.parser.sev_out_dict)} Severity(ies)')
            if len(rv) == 1:
                rv = rv.popitem()[1]

        # return created objects
        return rv

    def write_from_file(self, file_name, **kwargs):
        """
        read program from file. delegates to write

        :param file_name:
        :param update:
        :param verbose:
        :param log2:
        :param bs:
        :param kwargs:
        :return:
        """
        with open(file_name, 'r', encoding='utf-8') as f:
            portfolio_program = f.read()
        return self.write(portfolio_program, **kwargs)

    def write_test(self, portfolio_program):
        """
        write programs in testing mode

        dictionary definitions are added to uw but no objects are created

        returns data frame description of added severity/aggregate/portfolios

        the dataframe of aggregates can be used to create a portfolio (with all the aggregates) by calling

        ```Portfolio.from_DataFrame(name df)```

        TODO rationalize with parse_portfolio_program

        :param portfolio_program:
        :return: dictionary with keys sev agg port and assoicated dataframes
        """
        print('write_test deprecated...use parse_portfolio_porgram with output="dict".')
        raise RuntimeError
        # TODO once sure you don't need this delete!
        # logger.info(f'Runner.write_test | Executing program\n{portfolio_program[:500]}\n\n')
        # self._runner(portfolio_program)
        # ans = {}
        # if len(self.parser.sev_out_dict) > 0:
        #     for v in self.parser.sev_out_dict.values():
        #         Underwriter._add_defaults(v, 'sev')
        #     ans['sev'] = pd.DataFrame(list(self.parser.sev_out_dict.values()), index=self.parser.sev_out_dict.keys())
        # if len(self.parser.agg_out_dict) > 0:
        #     for v in self.parser.agg_out_dict.values():
        #         Underwriter._add_defaults(v)
        #     ans['agg'] = pd.DataFrame(list(self.parser.agg_out_dict.values()), index=self.parser.agg_out_dict.keys())
        # if len(self.parser.port_out_dict) > 0:
        #     ans['port'] = pd.DataFrame(list(self.parser.port_out_dict.values()), index=self.parser.port_out_dict.keys())
        # return ans

    def _runner(self, portfolio_program):
        """
        preprocessing:
            remove \n in [] (vectors) e.g. put by f{np.linspace} TODO only works for 1d vectors
            ; mapped to newline
            backslash (line continuation) mapped to space
            split on newlines
            parse one line at a time
            PIPE format no longer supported

        error handling and piping through parser

        :param portfolio_program:
        :return:
        """
        # Preprocess ---------------------------------------------------------------------
        # handle \n in vectors; first item is outside, then inside... (multidimensional??)
        # remove coments # xxx
        portfolio_program = re.sub(r'\s*#[^\n]*\n', r'\n', portfolio_program)
        out_in = re.split(r'\[|\]', portfolio_program)
        assert len(out_in) % 2  # must be odd
        odd = [t.replace('\n', ' ') for t in out_in[1::2]]  # replace inside []
        even = out_in[0::2]  # otherwise pass through
        # reassemble
        portfolio_program = ' '.join([even[0]] + [f'[{o}] {e}' for o, e in zip(odd, even[1:])])
        # other preprocessing: line continuation; \n\t or \n____ to space (for port indents),
        # ; to new line, split on new line
        portfolio_program = [i.strip() for i in portfolio_program.replace('\\\n', ' ').
            replace('\n\t', ' ').replace('\n    ', ' ').replace(';', '\n').
            split('\n') if len(i.strip()) > 0]

        # Parse      ---------------------------------------------------------------------
        self.parser.reset()
        for program_line in portfolio_program:
            # print(program_line)
            try:
                if len(program_line) > 0:
                    self.parser.parse(self.lexer.tokenize(program_line))
            except ValueError as e:
                if isinstance(e.args[0], str):
                    print(e)
                    raise e
                else:
                    t = e.args[0].type
                    v = e.args[0].value
                    i = e.args[0].index
                    txt2 = program_line[0:i] + f'>>>' + program_line[i:]
                    print(f'Parse error in input "{txt2}"\nValue {v} of type {t} not expected')
                    raise e

        # Post process -------------------------------------------------------------------
        if self.store_mode:
            # could do this with a loop and getattr but it is too hard to read, so go easy route
            if len(self.parser.sev_out_dict) > 0:
                # for k, v in self.parser.sev_out_dict.items():
                self.severity.update(self.parser.sev_out_dict)  # [k] = v
                logger.info(f'Underwriter._runner | saving {self.parser.sev_out_dict.keys()} severity/ies')
            if len(self.parser.agg_out_dict) > 0:
                # for k, v in self.parser.agg_out_dict.items():
                #     self.aggregate[k] = v
                self.aggregate.update(self.parser.agg_out_dict)
                logger.info(f'Underwriter._runner | saving {self.parser.agg_out_dict.keys()} aggregate(s)')
            if len(self.parser.port_out_dict) > 0:
                for k, v in self.parser.port_out_dict.items():
                    # v is a list of aggregate names, these have all been added to the database...
                    logger.info(f'Underwriter._runner | saving {k} portfolio')
                    self.portfolio[k] = {'spec': v['spec'], 'arg_dict': {}}
                    # self.portfolio[k] = {'spec': [self.aggregate[_a] for _a in v], 'arg_dict': {}}
        # can we still do something like this?
        #     self.parser.arg_dict['note'] = txt
        return

    @staticmethod
    def _add_defaults(dict_, kind='agg'):
        """
        add default values to dict_ Leave existing values unchanged

        :param dict_:
        :return:
        """
        defaults = dict(name="", exp_el=0, exp_premium=0, exp_lr=0, exp_en=0, exp_attachment=0, exp_limit=np.inf,
                        sev_name='', sev_a=0, sev_b=0, sev_mean=0, sev_cv=0, sev_scale=0, sev_loc=0, sev_wt=1,
                        freq_name='poisson', freq_a=0, freq_b=0)
        if kind == 'agg':
            for k, v in defaults.items():
                if k not in dict_:
                    dict_[k] = v
        elif kind == 'sev':
            for k, v in defaults.items():
                if k[0:3] == 'sev' and k not in dict_ and k != 'sev_wt':
                    dict_[k] = v

    def _safe_lookup(self, full_uw_id):
        """
        lookup uw_id in uw of expected type and merge safely into self.arg_dict
        delete name and note if appropriate

        :param full_uw_id:  type.name format
        :return:
        """

        expected_type, uw_id = full_uw_id.split('.')
        try:
            # lookup in Underwriter
            found_type, found_dict = self[uw_id]
        except LookupError as e:
            print(f'ERROR id {expected_type}.{uw_id} not found')
            raise e
        logger.info(f'UnderwritingParser._safe_lookup | retrieved {uw_id} as type {found_type}')
        if found_type != expected_type:
            raise ValueError(f'Error: type of {uw_id} is  {found_type}, not expected {expected_type}')
        return found_dict.copy()

    @staticmethod
    def obj_to_agg(obj):
        """
        convert an object into an agg language specification, used for saving
        :param obj: a dictionary, Aggregate, Severity or Portfolio object
        :return:
        """
        pass

# def dict_2_string(type_name, dict_in, tab_level=0, sio=None):
#     """
#     nice formating for str function
#
#     :param type_name:
#     :param dict_in:
#     :param tab_level:
#     :param sio:
#     :return:
#     """
#
#     if sio is None:
#         sio = StringIO()
#
#     keys = sorted(dict_in.keys())
#     if 'name' in keys:
#         # which it should always be
#         nm = dict_in['name']
#         sio.write(nm + '\n' + '=' * len(nm) + '\n')
#         keys.pop(keys.index('name'))
#
#     sio.write(f'{"type":<20s}{type_name}\n')
#
#     for k in keys:
#         v = dict_in[k]
#         ks = '\t' * max(0, tab_level - 1) + f'{str(k):<20s}'
#         if type(v) == dict:
#             # sio.write('\t' * tab_level + ks + '\n')
#             dict_2_string(type(v), v, tab_level + 1, sio)
#         elif isinstance(v, str):
#             if len(v) > 30:
#                 sio.write('\t' * tab_level + ks + '\n' + indent(fill(v, 30), ' ' * (4 * tab_level + 20)))
#             elif len(v) > 0:
#                 sio.write('\t' * tab_level + ks + v)
#             sio.write('\n')
#         elif isinstance(v, collections.Iterable):
#             sio.write('\t' * tab_level + ks + '\n')
#             for vv in v:
#                 sio.write('\t' * (tab_level + 1) + str(vv) + '\n')
#         elif type(v) == int:
#             sio.write('\t' * tab_level + f'{ks}\t{v:20d}\n')
#         elif type(v) == float:
#             if abs(v) < 100:
#                 sio.write('\t' * tab_level + f'{ks}\t{v:20.5f}\n')
#             else:
#                 sio.write('\t' * tab_level + f'{ks}\t{v:20,.1f}\n')
#         else:
#             # logger.info(f'Uknown type {type(v)} to dict_2_string')
#             sio.write('\t' * tab_level + ks + '\t' + str(v) + '\n')
#     return sio.getvalue()



"""
lexer and parser specification for aggregate
============================================

Overview
    Implements the ``agg`` programming lanaguage

Example program
    The following short program replicates Thought Experiemnt 1 from Neil Bodoff's
    paper Capital Allocation by Percentile Layer

    ::

        port BODOFF1 note{Bodoff Thought Experiment No. 1}
            agg wind  1 claim sev dhistogram xps [0,  99] [0.80, 0.20] fixed
            agg quake 1 claim sev dhistogram xps [0, 100] [0.95, 0.05] fixed

Preprocessing
    tab or four spaces needed and are replaced with space (program is on one line)


Ignored characters
    colon, comma, ( ) |


Language Specification
----------------------

The ```agg``` Language Grammar:

::

    answer              	:: sev_out
                        	 | agg_out
                        	 | port_out

    port_out            	:: port_name note agg_list

    agg_list            	:: agg_list agg_out
                        	 | agg_out

    agg_out             	:: agg_name builtin_aggregate note
                        	 | agg_name exposures layers SEV sev freq note

    sev_out             	:: sev_out sev_name sev note
                        	 | sev_name sev note

    freq                	:: MIXED ID snumber snumber
                        	 | MIXED ID snumber
                        	 | FREQ snumber
                        	 | FREQ

    snumber             	:: NUMBER
                        	 | MINUS NUMBER %prec UMINUS

    sev                 	:: sev UNCONDITIONAL
                             | sev PLUS numbers
                        	 | sev MINUS numbers
                        	 | numbers TIMES sev
                        	 | ids numbers CV numbers weights
                        	 | ids numbers weights
                        	 | ids numbers numbers weights xps
                        	 | ids xps
                        	 | builtinids numbers numbers
                        	 | builtinids

    xps                 	:: XPS numbers numbers
                        	 |

    weights             	:: WEIGHTS EQUAL_WEIGHT NUMBER
                        	 | WEIGHTS numbers
                        	 |

    layers              	:: numbers XS numbers
                        	 |

    note                	:: NOTE
                        	 |

    exposures           	:: numbers CLAIMS
                        	 | numbers LOSS
                        	 | numbers PREMIUM AT numbers LR
                        	 | numbers PREMIUM AT numbers

    builtinids          	:: BUILTINID

    ids                 	:: "[" idl "]"
                        	 | ID

    idl                 	:: idl ID
                        	 | ID

    numbers             	:: "[" numberl "]"
                        	 | NUMBER

    numberl             	:: numberl NUMBER
                        	 | NUMBER

    builtin_aggregate   	:: builtin_aggregate_dist TIMES NUMBER
                        	 | NUMBER TIMES builtin_aggregate_dist
                        	 | builtin_aggregate_dist

    builtin_aggregate_dist	:: BUILTINID

    sev_name            	:: SEV ID

    agg_name            	:: AGG ID

    port_name           	:: PORT ID

parser.out parser debug information
-----------------------------------

Lexer term definition
---------------------

::

    tokens = {ID, BUILTINID, NOTE,
              SEV, AGG, PORT,
              PLUS, MINUS, TIMES, NUMBER,
              LOSS, PREMIUM, AT, LR, CLAIMS,
              XS,
              CV, WEIGHTS, EQUAL_WEIGHT, XPS,
              MIXED, FREQ
              }
    ignore = ' \t,\\:\\(\\)|'
    literals = {'[', ']'}

    # per manual, need to list longer tokens before shorter ones
    # NOTE = r'note\{[0-9a-zA-Z,\.\(\)\-=\+!\s]*\}'  # r'[^\}]+'
    NOTE = r'note\{[^\}]*\}'  # r'[^\}]+'
    BUILTINID = r'(sev|agg|port|meta)\.[a-zA-Z][a-zA-Z0-9_]*'
    FREQ = r'binomial|poisson|bernoulli|fixed'
    ID = r'[a-zA-Z][\.a-zA-Z0-9~]*'  # do not allow _ in line names, use ~ instead: BECAUSE p_ etc. makes _ special
    PLUS = r'\+'
    MINUS = r'\-'
    TIMES = r'\*'
    EQUAL_WEIGHT = r'='
    ID['loss'] = LOSS
    ID['at'] = AT
    ID['cv'] = CV
    ID['premium'] = PREMIUM
    ID['prem'] = PREMIUM
    ID['lr'] = LR
    ID['claims'] = CLAIMS
    ID['claim'] = CLAIMS
    ID['x'] = XS
    ID['xs'] = XS
    ID['wts'] = WEIGHTS
    ID['wt'] = WEIGHTS
    ID['xps'] = XPS
    ID['mixed'] = MIXED
    ID['inf'] = NUMBER
    ID['sev'] = SEV
    ID['on'] = SEV
    ID['agg'] = AGG
    ID['port'] = PORT


Example Code
------------

    ::

        port Complex~Portfolio~Mixed
            agg LineA  50  claims           sev lognorm 12 cv [2, 3, 4] wt [.3 .5 .2] mixed gamma 0.4
            agg LineB  24  claims 10 x 5    sev lognorm 12 cv [{', '.join([str(i) for i in np.linspace(2,5, 20)])}] wt=20 mixed gamma 0.35
            agg LineC 124  claims 120 x 5   sev lognorm 16 cv 3.4                     mixed gamma 0.45

        port Complex~Portfolio
            agg Line3  50  claims [5 10 15] x 0         sev lognorm 12 cv [1, 2, 3]        mixed gamma 0.25
            agg Line9  24  claims [5 10 15] x 5         sev lognorm 12 cv [1, 2, 3] wt=3   mixed gamma 0.25

        port Portfolio~2
            agg CA 500 prem at .5 lr 15 x 12  sev gamma 12 cv [2 3 4] wt [.3 .5 .2] mixed gamma 0.4
            agg FL 1.7 claims 100 x 5         sev 10000 * pareto 1.3 - 10000        poisson
            agg IL 1e-8 * agg.CMP
            agg OH agg.CMP * 1e-8
            agg NY 500 prem at .5 lr 15 x 12  sev [20 30 40 10] * gamma [9 10 11 12] cv [1 2 3 4] wt =4 mixed gamma 0.4

        sev proda 30000 * lognorm 2
        sev prodc:   50000 * lognorm(3)
        sev weird    50000 * beta(1, 4) + 10000
        sev premsop1 25000 * lognorm 2.3; sev premsop2 35000 * lognorm 2.4;
        sev premsop3 45000 * lognorm 2.8

        agg Agg1     20 claims 10 x 2 sev lognorm 12 cv 0.2 mixed gamma 0.8
        agg Agg2     20 claims 10 x 2 sev 15 * lognorm 2.5  poisson;
        sev premsop1 25000 * lognorm 2.3;
        agg Agg3     20 claims 10 x 2 on 25 * lognorm 2 fixed;

        port MyFirstPortfolio
            agg A1: 50  claims          sev gamma 12 cv .30 (mixed gamma 0.014)
            agg A2: 50  claims 30 xs 10 sev gamma 12 cv .30 (mixed gamma 0.014)
            agg A3: 50  claims          sev gamma 12 cv 1.30 (mixed gamma 0.014)
            agg A4: 50  claims 30 xs 20 sev gamma 12 cv 1.30 (mixed gamma 0.14)
            agg B 15 claims 15 xs 15 sev lognorm 12 cv 1.5 + 2 mixed gamma 4.8
            agg Cat 1.7 claims 25 xs 5  sev 25 * pareto 1.3 0 - 25 poisson
            agg ppa: 1e-8 * agg.PPAL

        port distortionTest
            agg mix    50 claims              [50, 100, 150, 200] xs 0  sev lognorm 12 cv [1,2,3,4]    poisson
            agg low    500 premium at 0.5     5 xs 5                    sev gamma 12 cv .30            mixed gamma 0.2
            agg med    500 premium at 0.5 lr  15 xs 10                  sev gamma 12 cv .30            mixed gamma 0.4
            agg xsa    50  claims             30 xs 10                  sev gamma 12 cv .30            mixed gamma 1.2
            agg hcmp   1e-8 * agg.CMP
            agg ihmp   agg.PPAL * 1e-8

References
----------

https://sly.readthedocs.io/en/latest/sly.html

"""
from sly import Lexer, Parser
import logging
import numpy as np
import warnings

logger = logging.getLogger('aggregate')


class UnderwritingLexer(Lexer):

    tokens = {ID, BUILTINID, NOTE,
              SEV, AGG, PORT,
              PLUS, MINUS, TIMES, NUMBER,
              LOSS, PREMIUM, AT, LR, CLAIMS,
              XS,
              CV, WEIGHTS, EQUAL_WEIGHT, XPS,
              MIXED, FREQ
              }
    ignore = ' \t,\\:\\(\\)|'
    literals = {'[', ']', '!'}

    # per manual, need to list longer tokens before shorter ones
    # NOTE = r'note\{[0-9a-zA-Z,\.\(\)\-=\+!\s]*\}'  # r'[^\}]+'
    NOTE = r'note\{[^\}]*\}'  # r'[^\}]+'
    BUILTINID = r'(sev|agg|port|meta)\.[a-zA-Z][a-zA-Z0-9_]*'
    FREQ = r'binomial|poisson|bernoulli|pascal|geometric|fixed'
    # do not allow _ in line names, use ~ or . or : instead: why: because p_ is used and _ is special
    # on honor system...really need two types of ID, it is OK in a portfolio name
    ID = r'[a-zA-Z][\.:~_a-zA-Z0-9]*'
    PLUS = r'\+'
    MINUS = r'\-'
    TIMES = r'\*'
    EQUAL_WEIGHT = r'='
    ID['loss'] = LOSS
    ID['at'] = AT
    ID['cv'] = CV
    ID['premium'] = PREMIUM
    ID['prem'] = PREMIUM
    ID['lr'] = LR
    ID['claims'] = CLAIMS
    ID['claim'] = CLAIMS
    ID['x'] = XS
    ID['xs'] = XS
    ID['wts'] = WEIGHTS
    ID['wt'] = WEIGHTS
    ID['xps'] = XPS
    ID['mixed'] = MIXED
    ID['inf'] = NUMBER
    ID['sev'] = SEV
    # ID['on'] = SEV
    ID['agg'] = AGG
    ID['port'] = PORT
    # ID['!'] = UNCONDITIONAL

    @_(r'\-?(\d+\.?\d*|\d*\.\d+)([eE](\+|\-)?\d+)?')
    def NUMBER(self, t):
        if t.value == 'INF' or t.value == 'inf':
            t.value = np.inf
        else:
            t.value = float(t.value)
        return t

    @_(r'\n+')
    def newline(self, t):
        self.lineno += t.value.count('\n')

    def error(self, t):
        print(f"Illegal character '{t.value[0]:s}'")
        self.index += 1


class UnderwritingParser(Parser):
    # debugfile = 'parser.out'
    tokens = UnderwritingLexer.tokens
    precedence = (('left', PLUS, MINUS), ('left', TIMES), ('right', UMINUS))

    def __init__(self, safe_lookup_function, debug=False):
        self.arg_dict = None
        self.sev_out_dict = None
        self.agg_out_dict = None
        self.port_out_dict = None
        self.reset()
        # instance of uw class to look up severities
        self._safe_lookup = safe_lookup_function
        if debug:
            def _print(message):
                print(message)

            self.log = _print
        else:
            def _print(message):
                logger.info('UnderwritingParser | ' + message)

            self.log = _print

    def reset(self):
        # TODO Add sev_xs and sev_ps !!
        self.sev_out_dict = {}
        self.agg_out_dict = {}
        self.port_out_dict = {}

    @staticmethod
    def _check_vectorizable(value):
        """
        check the if value can be vectorized

        """
        if isinstance(value, float) or isinstance(value, int) or isinstance(value, np.ndarray):
            return value
        else:
            return np.array(value)

    # @staticmethod
    # def new_arg_dict():
    #     # to
    #     # in order to allow for missing terms this must reflect sensible defaults
    #     return dict(name="", exp_el=0, exp_premium=0, exp_lr=0, exp_en=0, exp_attachment=0, exp_limit=np.inf,
    #                 sev_name='', sev_a=0, sev_b=0, sev_mean=0, sev_cv=0, sev_scale=0, sev_loc=0, sev_wt=1,
    #                 freq_name='poisson', freq_a=0, freq_b=0)

    # final answer exit points ===================================
    @_('sev_out')
    def answer(self, p):
        self.log(f'\t\tExiting through sev_out, created severity {p.sev_out}')

    @_('agg_out')
    def answer(self, p):
        self.log(f'\t\tExiting through agg_out, created aggregate {p.agg_out}')

    @_('port_out')
    def answer(self, p):
        self.log(f'\t\tExiting through port_out, created portfolio {p.port_out} '
                 f'with {len(self.port_out_dict[p.port_out]["spec"])} aggregates')

    # building portfolios =======================================
    @_('port_name note agg_list')
    def port_out(self, p):
        self.log(f'\tADDING port_name note ({p.note[:10]}...) agg_list to port_out {p.port_name}')
        self.port_out_dict[p.port_name] = {'spec': p.agg_list, 'note': p.note}
        return p.port_name

    @_('agg_list agg_out')
    def agg_list(self, p):
        if p.agg_list is None:
            raise ValueError('ODD agg list is empty')
        p.agg_list.append(p.agg_out)
        self.log(f'\tADDED agg_out {p.agg_out} to agg_list {p.agg_list}')
        return p.agg_list

    @_('agg_out')
    def agg_list(self, p):
        self.log(f'\tADDING agg_out {p.agg_out} to new agg_list')
        return [p.agg_out]

    # building aggregates ========================================
    @_('agg_name builtin_aggregate note')
    def agg_out(self, p):
        self.log(f'ADDING agg_name builtin_aggregate note {p.builtin_aggregate} to agg_out')
        if 'name' in p.builtin_aggregate:
            # otherwise will overwrite the agg name
            del p.builtin_aggregate['name']
        self.agg_out_dict[p.agg_name] = {'name': p.agg_name, **p.builtin_aggregate, 'note': p.note}
        return p.agg_name

    # standard spec expos [layers] sevs freq
    @_('agg_name exposures layers SEV sev freq note')
    def agg_out(self, p):
        self.log(f'ADDING agg_name exposures layers SEV sev freq note to agg_out {p.agg_name}')
        self.agg_out_dict[p.agg_name] = {'name': p.agg_name, **p.exposures, **p.layers, **p.sev,
                                         **p.freq, 'note': p.note}
        return p.agg_name

    # building severities ======================================
    @_('sev_out sev_name sev note')
    def sev_out(self, p):
        self.log(f'ADDING sev_out sev_name {p.sev_name} sev note to sev_out, appending to sev_out_dict')
        p.sev['note'] = p.note
        self.sev_out_dict[p.sev_name] = p.sev

    @_('sev_name sev note')
    def sev_out(self, p):
        self.log(f'ADDING sev_name sev note resolving to sev_part {p.sev_name}, adding to sev_out_dict')
        p.sev['note'] = p.note
        self.sev_out_dict[p.sev_name] = p.sev

    # frequency term ==========================================
    # for all frequency distributions claim count is determined by exposure / severity
    # only freq shape parameters need be entered
    # one and two parameter mixing distributions
    @_('MIXED ID snumber snumber')
    def freq(self, p):
        self.log(f'MIXED ID snumber snumber {p.ID}, {p[2]}, {p[3]} to two param freq, snumber.1=CV')
        return {'freq_name': p.ID, 'freq_a': p[2], 'freq_b': p[3]}

    @_('MIXED ID snumber')
    def freq(self, p):
        self.log(f'MIXED ID snumber {p.ID}, {p.snumber} to single param freq, snumber=CVs')
        return {'freq_name': p.ID, 'freq_a': p.snumber}

    @_('FREQ snumber snumber')
    def freq(self, p):
        self.log(f'resolving named frequency distribution {p.FREQ} parameters {p[1]}, {p[2]} to freq')
        if p.FREQ != 'pascal':
            warnings.warn(f'Illogical choice of frequency {p.FREQ}, expected pascal')
        return {'freq_name': p.FREQ, 'freq_a': p[1], 'freq_b': p[2]}

    # binomial p or TODO inflated poisson
    @_('FREQ snumber')
    def freq(self, p):
        self.log(f'resolving named frequency distribution {p.FREQ} parameter {p.snumber} to freq')
        if p.FREQ != 'binomial':
            warnings.warn(f'Illogical choice of frequency {p.FREQ}, expected binomial')
        return {'freq_name': p.FREQ, 'freq_a': p.snumber}

    @_('FREQ')
    def freq(self, p):
        self.log(f'resolving named frequency distribution {p.FREQ} to freq')
        if p.FREQ not in ('poisson', 'bernoulli', 'fixed', 'geometric'):
            warnings.warn(f'Illogical choice for FREQ {p.FREQ}, should be poisson, bernoulli or fixed')
        return {'freq_name': p.FREQ}

    @_('NUMBER')
    def snumber(self, p):
        self.log(f'NUMBER {p.NUMBER} to signed number')
        return p.NUMBER

    @_('MINUS NUMBER %prec UMINUS')
    def snumber(self, p):
        self.log(f'-NUMBER {p.NUMBER} to signed number')
        return -p.NUMBER

    # require a frequency distribution
    # @_('')
    # def freq(self, p):
    #     self.log('missing frequency term')
    #     return { 'freq_name': 'poisson'}

    # severity term ============================================
    @_('sev "!"')
    def sev(self, p):
        self.log(f'setting conditional flag on severity to False (this is unusual behaviour, '
                  'usually want conditional severity)')
        p.sev['sev_conditional'] = False
        return p.sev

    @_('sev PLUS numbers')
    def sev(self, p):
        self.log(f'resolving sev PLUS numbers to sev {p.numbers}')
        p.sev['sev_loc'] = UnderwritingParser._check_vectorizable(p.sev.get('sev_loc', 0))
        p.numbers = UnderwritingParser._check_vectorizable(p.numbers)
        p.sev['sev_loc'] += p.numbers
        return p.sev

    @_('sev MINUS numbers')
    def sev(self, p):
        self.log(f'resolving sev MINUS numbers to sev {p.numbers}')
        p.sev['sev_loc'] = p.sev.get('sev_loc', 0) - p.numbers
        return p.sev

    @_('numbers TIMES sev')
    def sev(self, p):
        self.log(f'resolving numbers TIMES sev to sev {p.numbers}')
        p.numbers = UnderwritingParser._check_vectorizable(p.numbers)
        if 'sev_mean' in p.sev:
            p.sev['sev_mean'] = UnderwritingParser._check_vectorizable(p.sev.get('sev_mean', 0))
            p.sev['sev_mean'] *= p.numbers
        # only scale scale if there is a scale (otherwise you double count)
        # TODO OK? sev_scale...
        if 'sev_scale' in p.sev:
            p.sev['sev_scale'] = UnderwritingParser._check_vectorizable(p.sev.get('sev_scale', 0))
            p.sev['sev_scale'] *= p.numbers
        if 'sev_mean' not in p.sev:
            # e.g. Pareto has no mean and it is important to set the scale
            # but if there is a mean it handles the scaling and setting scale will
            # confuse the distribution maker
            p.sev['sev_scale'] = p.numbers
        # if there is a location it needs to scale too --- that's a curious choice!
        if 'sev_loc' in p.sev:
            p.sev['sev_loc'] = UnderwritingParser._check_vectorizable(p.sev['sev_loc'])
            p.sev['sev_loc'] *= p.numbers
        return p.sev

    @_('ids numbers CV numbers weights')
    def sev(self, p):
        self.log(f'resolving ids {p.ids} numbers CV numbers weights {p[1]}, {p[3]}, {p.weights}')
        return {'sev_name':  p.ids, 'sev_mean':  p[1], 'sev_cv':  p[3], 'sev_wt': p.weights}

    @_('ids numbers weights')
    def sev(self, p):
        self.log(f'resolving ids {p.ids} numbers {p[1]} to sev (one param dist)')
        return {'sev_name': p.ids, 'sev_a':  p[1], 'sev_wt': p.weights}

    #                                v can go
    #@_('ids numbers numbers weights xps')
    @_('ids numbers numbers weights')
    def sev(self, p):
        self.log(f'resolving ids {p.ids} numbers numbers {p[1]}, {p[2]} to sev (two param sev dist)')
        # return {'sev_name': p.ids, 'sev_a':  p[1], 'sev_b':  p[2], 'sev_wt': p.weights, **p.xps}
        return {'sev_name': p.ids, 'sev_a': p[1], 'sev_b': p[2], 'sev_wt': p.weights}

    # TODO a bit restrictive on numerical densities here!
    #      v put in weights here instead (if xps relevant then cannot need shape parameters)
    @_('ids xps')
    def sev(self, p):
        self.log(f'resolving ids {p.ids} xps {p.xps} to sev (fixed or histogram type)')
        return {'sev_name': p.ids, **p.xps}

    @_('XPS numbers numbers')
    def xps(self, p):
        self.log(f'XPS numbers numbers resolving to xs and ps {p[1]}, {p[2]}')
        return {'sev_xs':  p[1], 'sev_ps':  p[2]}

    @_('')
    def xps(self, p):
        self.log('missing xps term')
        return {}

    @_('WEIGHTS EQUAL_WEIGHT NUMBER')
    def weights(self, p):
        self.log(f'WEIGHTS EQUAL_WEIGHTS {p.NUMBER} resolving to equal weights')
        return np.ones(int(p.NUMBER)) / p.NUMBER

    @_('WEIGHTS numbers')
    def weights(self, p):
        self.log(f'WEIGHTS numbers resolving to weights {p.numbers}')
        return p.numbers

    @_('')
    def weights(self, p):
        self.log('missing weights term')
        return 1

    @_('builtinids numbers numbers')
    def sev(self, p):
        self.log(f'builtinds {p.builtinids} numbers numbers log2={p[1]}, bs={p[2]} to sev')
        requested_type = p.builtinids.split('.')[0]
        if requested_type == "meta":
            return {'sev_name': p.builtinids, 'sev_a': p[1], 'sev_b': p[2]}
        else:
            raise ValueError(f'Only meta type can be used with arguments, not {p.builtinids}')

    @_('builtinids')
    def sev(self, p):
        self.log(f'builtinds {p.builtinids} to sev')
        # look up ID in uw
        # it is not accepetable to ask for an agg or port here; they need to be accessed through
        # meta. E.g. if you request and agg it will overwrite other (freq) variables defined
        # in the script...
        requested_type = p.builtinids.split('.')[0]
        if requested_type not in ("sev", "meta"):
            raise ValueError(f'built in type must be sev or meta, not {p.builtinids}')
        if requested_type == 'meta':
            return {'sev_name': p.builtinids}
        else:
            return self._safe_lookup(p.builtinids)

    # layer terms, optoinal ===================================
    @_('numbers XS numbers')
    def layers(self, p):
        self.log(f'resolving numbers XS numbers to layers {p[0]} xs {p[2]}')
        return {'exp_attachment': p[2], 'exp_limit': p[0]}

    @_('')
    def layers(self, p):
        self.log('missing layer term')
        return {}

    # optional note  ==========================================
    @_('NOTE')
    def note(self, p):
        self.log(f'NOTE to note: {p.NOTE[5:-1]}')
        return p.NOTE[5:-1]

    @_("")
    def note(self, p):
        self.log("Empty note term")
        return ''

    # exposures term ==========================================
    @_('numbers CLAIMS')
    def exposures(self, p):
        self.log(f'resolving numbers CLAIMS to exposures {p.numbers}')
        return {'exp_en': p.numbers}

    @_('numbers LOSS')
    def exposures(self, p):
        self.log(f'resolving numbers LOSS to exposures {p.numbers}')
        return {'exp_el': p.numbers}

    @_('numbers PREMIUM AT numbers LR')
    def exposures(self, p):
        self.log(f'resolving numbers PREMIUM AT numbers LR to exposures {p[0]} at {p[3]}')
        return {'exp_premium': p[0], 'exp_lr': p[3], 'exp_el': p[0] * p[3]}

    @_('numbers PREMIUM AT numbers')
    def exposures(self, p):
        self.log(f'resolving numbers PREMIUM AT numbers to exposures {p[0]} at {p[3]}')
        return {'exp_premium': p[0], 'exp_lr': p[3], 'exp_el': p[0] * p[3]}

    # lists for ids and numbers and builtinids ================================
    # for now, do not allow a list of severities...too tricky
    # @_('"[" builtinidl "]"')
    # def builtinids(self, log):
    #     self.log(f'resolving [builtinidl] to builtinids {p.builtinidl}')
    #     return p.builtinidl
    #
    # @_('builtinidl BUILTINID')
    # def builtinidl(self, log):
    #     s1 = f'resolving builtinidl BUILTINID {p.builtinidl}, {p.BUILTINID} --> '
    #     p.builtinidl.append(p.BUILTINID)
    #     s1 += f'{p.builtinidl}'
    #     self.log(s1)
    #     return p.builtinidl
    #
    # @_('BUILTINID')
    # def builtinidl(self, log):
    #     self.log(f'resolving BUILTINID to builtinidl {p.BUILTINID} --> {ans}')
    #     ans = [p.BUILTINID]
    #     return ans

    @_('BUILTINID')
    def builtinids(self, p):
        self.log(f'resolving BUILTINID to builtinids {p.BUILTINID}')
        return p.BUILTINID  # will always be treated as a list

    @_('"[" idl "]"')
    def ids(self, p):
        self.log(f'resolving [id1] to ids {p.idl}')
        return p.idl

    @_('idl ID')
    def idl(self, p):
        s1 = f'resolving idl ID {p.idl}, {p.ID} --> '
        p.idl.append(p.ID)
        s1 += f'{p.idl}'
        self.log(s1)
        return p.idl

    @_('ID')
    def idl(self, p):
        ans = [p.ID]
        self.log(f'resolving ID to idl {p.ID} --> {ans}')
        return ans

    @_('ID')
    def ids(self, p):
        self.log(f'resolving ID to ids {p.ID}')
        return p.ID

    @_('"[" numberl "]"')
    def numbers(self, p):
        self.log(f'resolving [number1] to numbers {p.numberl}')
        return p.numberl

    @_('numberl NUMBER')
    def numberl(self, p):
        s1 = f'resolving numberl NUMBER {p.numberl}, {p.NUMBER} --> '
        p.numberl.append(p.NUMBER)
        s1 += f'{p.numberl}'
        self.log(s1)
        return p.numberl

    @_('NUMBER')
    def numberl(self, p):
        ans = [p.NUMBER]
        self.log(f'resolving NUMBER to numberl {p.NUMBER} --> {ans}')
        return ans

    @_('NUMBER')
    def numbers(self, p):
        self.log(f'resolving NUMBER to numbers {p.NUMBER}')
        return p.NUMBER

    # elements made from named portfolios ========================
    @_('builtin_aggregate_dist TIMES NUMBER')
    def builtin_aggregate(self, p):
        """  inhomogeneous change of scale """
        self.log(f'builtin_aggregate_dist TIMES NUMBER {p.NUMBER}')
        bid = p.builtin_aggregate_dist
        bid['exp_en'] = bid.get('exp_en', 0) * p.NUMBER
        bid['exp_el'] = bid.get('exp_el', 0) * p.NUMBER
        bid['exp_premium'] = bid.get('exp_premium', 0) * p.NUMBER
        return bid

    @_('NUMBER TIMES builtin_aggregate_dist')
    def builtin_aggregate(self, p):
        """
        homogeneous change of scale

        :param p:
        :return:
        """
        self.log(f'NUMBER {p.NUMBER} TIMES builtin_aggregate_dist')
        # bid = built_in_dict, want to be careful not to add scale too much
        bid = p.builtin_aggregate_dist  # ? does this need copying. if so do in safelookup!
        if 'sev_mean' in bid:
            bid['sev_mean'] = bid['sev_mean'] * p.NUMBER
        if 'sev_scale' in bid:
            bid['sev_scale'] = bid['sev_scale'] * p.NUMBER
        if 'sev_loc' in bid:
            bid['sev_loc'] = bid['sev_loc'] * p.NUMBER
        bid['exp_attachment'] = bid.get('exp_attachment', 0) * p.NUMBER
        bid['exp_limit'] = bid.get('exp_limit', np.inf) *p.NUMBER
        bid['exp_el'] = bid.get('exp_el', 0) * p.NUMBER
        bid['exp_premium'] = bid.get('exp_premium', 0) * p.NUMBER
        return bid

    @_('builtin_aggregate_dist')
    def builtin_aggregate(self, p):
        self.log('builtin_aggregate_dist becomese builtin_aggregate')
        return p.builtin_aggregate_dist

    @_('BUILTINID')
    def builtin_aggregate_dist(self, p):
        # ensure lookup only happens here
        self.log(f'Lookup BUILTINID {p.BUILTINID}')
        built_in_dict = self._safe_lookup(p.BUILTINID)
        return built_in_dict

    # ids =========================================================
    @_('SEV ID')
    def sev_name(self, p):
        self.log(f'SEV ID resolves to sev_name {p.ID}')
        return p.ID

    @_('AGG ID')
    def agg_name(self, p):
        self.log(f'AGG ID resolves to agg_name {p.ID}')
        # return {'name': p.ID}
        return p.ID

    # require the AGG keyword to start a new agg
    # @_('ID')
    # def agg_name(self, p):
    #     self.log(f'ID resolves to agg_name {p.ID}')
    #     # return {'name': p.ID}
    #     return p.ID

    @_('PORT ID')
    def port_name(self, p):
        self.log(f'PORT ID note resolves to port_name {p.ID}')
        # return {'name': p.ID}
        return p.ID

    def error(self, p):
        if p:
            raise ValueError(p)
        else:
            raise ValueError('Unexpected end of file')


if __name__ == '__main__':
    # print the grammar and add to this file as part of docstring
    # TODO fix comments!

    # may need to put an extra indent for rst to work properly

    start_string = '''Language Specification
----------------------

The ```agg``` Language Grammar:

::

'''
    end_string = 'parser.out parser debug information'

    with open(__file__, 'r') as f:
        txt = f.read()
    stxt = txt.split('@_')
    ans = {}
    for it in stxt[3:-2]:
        if it.find('# def') >= 0:
            # skip rows with a comment between @_ and def
            pass
        else:
            b = it.split('def')
            b0 = b[0].strip()[2:-2]
            b1 = b[1].split("(self, p):")[0].strip()
            if b1 in ans:
                ans[b1] += [b0]
            else:
                ans[b1] = [b0]
    s = ''
    for k, v in ans.items():
        s += f'{k:<20s}\t:: {v[0]:<s}\n'
        for rhs in v[1:]:
            s += f'{" "*20}\t | {rhs:<s}\n'
        s += '\n'
    print(s)
    st = txt.find(start_string) + len(start_string)
    end = txt.find(end_string)
    txt = txt[0:st] + s + txt[end:]
    with open(__file__, 'w') as f:
        f.write(txt)



# https://pythonhosted.org/an_example_pypi_project/setuptools.html
# set HOME=c:\s\telos\python
# python setup.py sdist bdist_wininst upload.bat
# pip freeze > file to list installed packages
# pip install -r requirements.txt to install

from setuptools import setup
import os
from time import time

tests_require = ['unittest', 'sly']
install_requires = [
    'pypandoc',
    'sly==0.3',
    'scipy',
    'seaborn',
    'matplotlib',
    'numpy'
]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()


long_description = read('README.rst')


def version_number():
    with open('version_number.md', 'r') as f:
        t = f.read()
    lt = t.split('\n')
    last_update = float(lt[1])
    # updates at least one hour apart
    if time() - last_update > 60 * 60:
        s = lt[0].strip().split('.')
        n = int(s[-1])
        s[-1] = str(n+1)
        t = '.'.join(s)
        with open('version_number.md', 'w') as f:
            f.write(f'{t}\n{time()}')
            print(f'Updated version number to {t}')
    else:
        t = lt[0].strip()
        print('No change')
    return t

# agg/*.agg is the parameter and curve files

setup(name="aggregate",
      description="aggregate - working with compound probability distributions",
      long_description=long_description,
      license="""BSD""",
      version=version_number(),
      author="Stephen J. Mildenhall",
      author_email="steve@convexrisk.com",
      maintainer="Stephen J. Mildenhall",
      maintainer_email="steve@convexrisk.com",
      packages=['aggregate'],
      package_data={'': ['*.txt', '*.rst', '*.md', 'agg/*.agg', 'examples/*.py', 'examples/*.ipynb',
                         'test/*.py']},
      tests_require=tests_require,
      install_requires=install_requires,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: BSD License',
          'Topic :: Education',
          'Topic :: Office/Business :: Financial',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Intended Audience :: Financial and Insurance Industry',
          'Intended Audience :: Education'
      ],
      project_urls={"Documentation": 'https://aggregate.readthedocs.io/en/latest/',
                    "Source Code": "https://github.com/mynl/aggregate_project"}
      )
